"""Metadata convert functions."""

from nativelib import Column
from pgcopylib import PGOid


def pgoid_from_metadata(metadata: bytes) -> list[PGOid]:
    """Convert PGPack metadata to PGCopy metadata."""

    ...


def columns_from_metadata(
    metadata: bytes,
    is_nullable: bool = True,
) -> list[Column]:
    """Convert PGPack metadata to Native column_list."""

    ...


def metadata_from_columns(column_list: list[Column]) -> bytes:
    """Convert Native column_list to PGPack metadata."""

    ...



"""Dump convert functions."""

from collections.abc import Generator
from enum import Enum
from io import BufferedReader
from itertools import chain
from typing import NamedTuple

from light_compressor import (
    CompressionMethod,
    define_reader,
    define_writer,
)
from nativelib import (
    NativeReader,
    NativeWriter,
)
from pgcopylib import (
    PGCopyReader,
    PGCopyWriter,
)
from pgpack import (
    PGPackReader,
    PGPackWriter,
)

from .common import (
    columns_from_metadata,
    metadata_from_columns,
    pgoid_from_metadata,
)
from .detective import dump_detective as detective


CHUNK_SIZE = 1_048_576


class DumpClass(NamedTuple):
    """Class for reader/writer implementations."""

    name: str
    reader: object
    writer: object
    have_compress: bool


class DumpType(DumpClass, Enum):
    """Dump type enum."""

    NATIVE = DumpClass("native", NativeReader, NativeWriter, False)
    PGCOPY = DumpClass("pgcopy", PGCopyReader, PGCopyWriter, False)
    PGPACK = DumpClass("pgpack", PGPackReader, PGPackWriter, True)


def chunk_fileobj(fileobj: BufferedReader) -> Generator[bytes, None, None]:
    """Chunk fileobject to bytes generator."""

    while chunk := fileobj.read(CHUNK_SIZE):
        yield chunk

    fileobj.close()


def dump_convertor(
    source: str,
    destination: str,
    dump_type: DumpType | str,
    compression_method: CompressionMethod | str = CompressionMethod.NONE,
) -> None:
    """Convert dumps function."""

    if dump_type.__class__ is str:
        dump_type = DumpType[dump_type.upper()]

    if compression_method.__class__ is str:
        compression_method = CompressionMethod[compression_method.upper()]

    reader: NativeReader | PGCopyReader | PGPackReader = detective(source)

    if reader.__class__ is PGCopyReader and dump_type is not DumpType.PGCOPY:
        raise TypeError(
            "PGCopy dump don't support convert to any other types."
        )
    if source == destination:
        raise ValueError("Files path is not different.")

    if reader.__class__ is dump_type.reader:
        if dump_type.reader in (NativeReader, PGCopyReader):
            reader.close()
            source_obj = define_reader(open(source, "rb"))
            writer = define_writer(
                chunk_fileobj(source_obj),
                compression_method,
            )

            with open(destination, "wb") as fileobj:
                for bytes_data in writer:
                    fileobj.write(bytes_data)

        elif dump_type.reader is PGPackReader:
            metadata = reader.metadata
            bytes_data = reader.to_bytes()
            fileobj = open(destination, "wb")
            writer = PGPackWriter(fileobj, metadata, compression_method)
            writer.from_bytes(bytes_data)
            writer.close()
    else:
        if reader.__class__ is NativeReader:
            dtype_values = chain(reader.block_reader.read(), reader.to_rows())
            column_list = reader.block_reader.column_list
            metadata = metadata_from_columns(column_list)
            fileobj = open(destination, "wb")

            if dump_type.writer is PGCopyWriter:
                pgtypes = pgoid_from_metadata(metadata)
                writer = PGCopyWriter(None, pgtypes)
                bytes_data = writer.from_rows(reader.to_rows())

                with fileobj:
                    for data in define_writer(
                        bytes_data,
                        compression_method,
                    ):
                        fileobj.write(data)

            elif dump_type.writer is PGPackWriter:
                writer = PGPackWriter(fileobj, metadata, compression_method)
                writer.from_rows(dtype_values)
                writer.close()
        elif reader.__class__ is PGPackReader:
            if dump_type.writer is NativeWriter:
                column_list = columns_from_metadata(reader.metadata)
                writer = NativeWriter(column_list)
                bytes_data = writer.from_rows(reader.to_rows())
            elif dump_type.writer is PGCopyWriter:
                bytes_data = reader.to_bytes()

            with open(destination, "wb") as fileobj:
                for data in define_writer(
                    bytes_data,
                    compression_method,
                ):
                    fileobj.write(data)

    reader.close()



"""Metadata convert functions."""

from json import dumps

from pgcopylib import PGOid
from pgcopylib.common import ArrayOidToOid
from pgpack import metadata_reader

from nativelib import Column


cdef dict DTYPE2OID = {
    "BFloat16": PGOid.float4,
    "Bool": PGOid.bool,
    "Date": PGOid.date,
    "Date32": PGOid.date,
    "DateTime": PGOid.timestamp,
    "DateTime64": PGOid.timestamptz,
    "Decimal": PGOid.numeric,
    "Enum8": PGOid.varchar,
    "Enum16": PGOid.varchar,
    "FixedString": PGOid.bpchar,
    "Float32": PGOid.float4,
    "Float64": PGOid.float8,
    "IPv4": PGOid.inet,
    "IPv6": PGOid.inet,
    "Int128": PGOid.int8,
    "Int16": PGOid.int2,
    "Int256": PGOid.int8,
    "Int32": PGOid.int4,
    "Int64": PGOid.int8,
    "Int8": PGOid.int2,
    "String": PGOid.varchar,
    "Time": PGOid.time,
    "Time64": PGOid.timetz,
    "UInt128": PGOid.int8,
    "UInt16": PGOid.int4,
    "UInt256": PGOid.int8,
    "UInt32": PGOid.int8,
    "UInt64": PGOid.int8,
    "UInt8": PGOid.int2,
    "UUID": PGOid.uuid,
}
cdef dict OID2ARRAY = {
    PGOid.bit: PGOid._bit,
    PGOid.bool: PGOid._bool,
    PGOid.box: PGOid._box,
    PGOid.bpchar: PGOid._bpchar,
    PGOid.bytea: PGOid._bytea,
    PGOid.char: PGOid._char,
    PGOid.cidr: PGOid._cidr,
    PGOid.circle: PGOid._circle,
    PGOid.date: PGOid._date,
    PGOid.float4: PGOid._float4,
    PGOid.float8: PGOid._float8,
    PGOid.inet: PGOid._inet,
    PGOid.int2: PGOid._int2,
    PGOid.int4: PGOid._int4,
    PGOid.int8: PGOid._int8,
    PGOid.interval: PGOid._interval,
    PGOid.json: PGOid._json,
    PGOid.jsonb: PGOid._jsonb,
    PGOid.line: PGOid._line,
    PGOid.lseg: PGOid._lseg,
    PGOid.macaddr: PGOid._macaddr,
    PGOid.macaddr8: PGOid._macaddr8,
    PGOid.money: PGOid._money,
    PGOid.numeric: PGOid._numeric,
    PGOid.oid: PGOid._oid,
    PGOid.path: PGOid._path,
    PGOid.point: PGOid._point,
    PGOid.polygon: PGOid._polygon,
    PGOid.text: PGOid._text,
    PGOid.time: PGOid._time,
    PGOid.timestamp: PGOid._timestamp,
    PGOid.timestamptz: PGOid._timestamptz,
    PGOid.timetz: PGOid._timetz,
    PGOid.uuid: PGOid._uuid,
    PGOid.varbit: PGOid._varbit,
    PGOid.varchar: PGOid._varchar,
    PGOid.xml: PGOid._xml,
}
cdef dict OID2DTYPE = {
    PGOid.bit: "String",
    PGOid.bool: "Bool",
    # PGOid.box: 603,
    PGOid.bpchar: "FixedString",
    # PGOid.bytea: 17,
    PGOid.char: "String",
    # PGOid.cidr: 650,
    # PGOid.circle: 718,
    PGOid.date: "Date",
    PGOid.float4: "Float32",
    PGOid.float8: "Float64",
    PGOid.inet: "IPv4",
    PGOid.int2: "Int16",
    PGOid.int4: "Int32",
    PGOid.int8: "Int64",
    PGOid.interval: "IntervalSecond",
    PGOid.json: "JSON",
    PGOid.jsonb: "JSON",
    # PGOid.line: 628,
    # PGOid.lseg: 601,
    PGOid.macaddr: "String",
    PGOid.macaddr8: "String",
    PGOid.money: "Float64",
    PGOid.numeric: "Decimal",
    PGOid.oid: "UInt32",
    # PGOid.path: 602,
    PGOid.point: "Point",
    PGOid.polygon: "Polygon",
    PGOid.text: "String",
    PGOid.time: "Time",
    PGOid.timestamp: "DateTime",
    PGOid.timestamptz: "DateTime64",
    PGOid.timetz: "Time64",
    PGOid.uuid: "UUID",
    PGOid.varbit: "String",
    PGOid.varchar: "String",
    PGOid.xml: "String",
}
cdef dict PGLENGTH = {
    PGOid.bool: 1,
    PGOid.box: 32,
    PGOid.circle: 24,
    PGOid.date: 4,
    PGOid.float4: 4,
    PGOid.float8: 8,
    PGOid.int2: 2,
    PGOid.int4: 4,
    PGOid.int8: 8,
    PGOid.interval: 16,
    PGOid.line: 24,
    PGOid.lseg: 32,
    PGOid.macaddr8: 8,
    PGOid.macaddr: 6,
    PGOid.oid: 4,
    PGOid.point: 16,
    PGOid.time: 8,
    PGOid.timestamp: 8,
    PGOid.timestamptz: 8,
    PGOid.timetz: 12,
    PGOid.uuid: 16,
}


cpdef list pgoid_from_metadata(bytes metadata):
    """Convert PGPack metadata to PGCopy metadata."""

    return metadata_reader(metadata)[1]


cpdef list columns_from_metadata(
    bytes metadata,
    object is_nullable = True,
):
    """Convert PGPack metadata to Native column_list."""

    cdef int _i
    cdef str column, dtype
    cdef object pgoid, pgparam
    cdef list column_list = []

    for column, pgoid, pgparam in zip(*metadata_reader(metadata)):

        if pgparam.nested:
            pgoid = ArrayOidToOid[pgoid]

        dtype = OID2DTYPE[pgoid]

        if dtype in ("DateTime64", "Time64"):
            dtype += "(3)"
        elif dtype == "Decimal":
            dtype += f"({pgparam.length}, {pgparam.scale})"
        elif dtype == "FixedString":
            dtype += f"({pgparam.length})"

        if is_nullable:
            dtype = f"Nullable({dtype})"

        for _i in range(pgparam.nested):
            dtype = f"Array({dtype})"

        column_list.append(Column(column, dtype))

    return column_list


cpdef bytes metadata_from_columns(list column_list):
    """Convert Native column_list to PGPack metadata."""

    cdef list pg_column_info, column_info, metadata = []
    cdef int number, lengths, scale, nested
    cdef object column, info, oid
    cdef str name, dtype, json_metadata

    for number, column in enumerate(column_list, 1):
        info = column.info
        name = info.column
        dtype = info.dtype.name
        oid = DTYPE2OID[dtype]

        if oid is PGOid.bpchar:
            lengths = info.length
        elif oid is PGOid.numeric:
            lengths = info.precission
        else:
            lengths = PGLENGTH.get(oid, -1)

        scale = info.scale or 0
        nested = info.nested

        if info.is_array:
            oid = OID2ARRAY[oid]

        pg_column_info = [name, oid.value, lengths, scale, nested]
        column_info = [number, pg_column_info]
        metadata.append(column_info)

    json_metadata = dumps(metadata, ensure_ascii=False)
    return json_metadata.encode("utf-8")



"""DBHose dumps convertors utility."""

from light_compressor import CompressionMethod

from .common import (
    columns_from_metadata,
    metadata_from_columns,
    pgoid_from_metadata,
)
from .convertor import (
    DumpClass,
    DumpType,
    chunk_fileobj,
    dump_convertor,
)
from .detective import dump_detective


__all__ = (
    "CompressionMethod",
    "DumpClass",
    "DumpType",
    "chunk_fileobj",
    "columns_from_metadata",
    "dump_convertor",
    "dump_detective",
    "metadata_from_columns",
    "pgoid_from_metadata",
)
__author__ = "0xMihalich"
__version__ = "0.0.0.4"



from setuptools import (
    Extension,
    setup,
)
from Cython.Build import cythonize


extensions = [
    Extension(
        "dbhose_utils.common",
        ["src/dbhose_utils/common.pyx"],
    ),
]


setup(
    name="dbhose_utils",
    package_dir={"": "src"},
    ext_modules=cythonize(extensions, language_level="3"),
    packages=[
        "dbhose_utils.common",
    ],
    package_data={
        "dbhose_utils.common": [
            "**/*.pyx",
            "**/*.pyi",
            "**/*.pxd",
            "*.pxd",
            "*.pyd",
            "*.md",
            "*.txt",
        ]
    },
    exclude_package_data={
        "": ["*.c"],
        "dbhose_utils.common": ["**/*.c"],
    },
    include_package_data=True,
    setup_requires=["Cython>=3.0"],
    zip_safe=False,
)



"""Auto detect dump type and compression method."""

from io import BufferedReader
from pathlib import Path

from light_compressor import define_reader
from nativelib import NativeReader
from pgcopylib import PGCopyReader
from pgpack import PGPackReader
from pgpack.common import HEADER


PGCOPY_HEADER = b"PGCOPY\n\xff"


def dump_detective(
    file: str | Path | BufferedReader
) -> PGCopyReader | PGPackReader | NativeReader:
    """Auto detect dump type and return reader."""

    if isinstance(file, str | Path):
        fileobj = open(file, "rb")
    elif hasattr(file, "read"):
        fileobj = file
    else:
        raise ValueError(
            "File type must be str, Path, BytesIO or "
            f"BufferedReader not {file.__class__}"
        )

    buffer_obj = define_reader(fileobj)
    pos = buffer_obj.tell()
    header = buffer_obj.read(8)
    buffer_obj.seek(pos)

    if header == HEADER:
        return PGPackReader(buffer_obj)
    if header == PGCOPY_HEADER:
        return PGCopyReader(buffer_obj)

    return NativeReader(buffer_obj)
